#include <Wire.h>
#include <FastLED.h>
#include <Control_Surface.h>

#include <AH/Hardware/MCP23017Encoders.hpp>

// Include the display interface you'd like to use
#include <Display/DisplayInterfaces/DisplayInterfaceSSD1306.hpp>


// ----------------------------- MIDI Interface ----------------------------- //
// ========================================================================== //
/*
   Instantiate a MIDI interface to use for the Control Surface.
*/

// Instantiate a MIDI over USB interface.
USBMIDI_Interface midi;

// ----------------------------- Display setup ------------------------------ //
// ========================================================================== //
/*
   Instantiate and initialize the SSD1306 OLED display
*/

constexpr uint8_t SCREEN_WIDTH = 128;
constexpr uint8_t SCREEN_HEIGHT = 64;

constexpr int8_t OLED_CLK_1_2 = 7;  //D0 pin on OLED Display PCB
constexpr int8_t OLED_CLK_3_4 = 7;
constexpr int8_t OLED_CLK_5_6 = 7;
constexpr int8_t OLED_CLK_7_8 = 7;

constexpr int8_t OLED_MOSI_1_2 = 8;  //D1 pin on OLED Display PCB
constexpr int8_t OLED_MOSI_3_4 = 8;
constexpr int8_t OLED_MOSI_5_6 = 8;
constexpr int8_t OLED_MOSI_7_8 = 8;

constexpr int8_t OLED_reset_1_2 = -1;  //RES (Reset) pin on OLED Display PCB
constexpr int8_t OLED_reset_3_4 = -1;
constexpr int8_t OLED_reset_5_6 = -1;
constexpr int8_t OLED_reset_7_8 = -1;

constexpr int8_t OLED_DC_1_2 = 10;    //DC (Data/Command) pin on OLED Display PCB
constexpr int8_t OLED_DC_3_4 = 10;
constexpr int8_t OLED_DC_5_6 = 10;
constexpr int8_t OLED_DC_7_8 = 10;

constexpr int8_t OLED_CS_1_2 = 11;    //CS (Chip Select) pin on OLED Display PCB
constexpr int8_t OLED_CS_3_4 = 12;
constexpr int8_t OLED_CS_5_6 = 9;
constexpr int8_t OLED_CS_7_8 = 6;

constexpr uint32_t SPI_Frequency = SPI_MAX_SPEED;


// Instantiate the displays
Adafruit_SSD1306 ssd1306Display_1_2(SCREEN_WIDTH, SCREEN_HEIGHT,
                                    OLED_MOSI_1_2, OLED_CLK_1_2, OLED_DC_1_2, OLED_reset_1_2, OLED_CS_1_2);

Adafruit_SSD1306 ssd1306Display_3_4(SCREEN_WIDTH, SCREEN_HEIGHT,
                                    OLED_MOSI_3_4, OLED_CLK_3_4, OLED_DC_3_4, OLED_reset_3_4, OLED_CS_3_4);

Adafruit_SSD1306 ssd1306Display_5_6(SCREEN_WIDTH, SCREEN_HEIGHT,
                                    OLED_MOSI_5_6, OLED_CLK_5_6, OLED_DC_5_6, OLED_reset_5_6, OLED_CS_5_6);

Adafruit_SSD1306 ssd1306Display_7_8(SCREEN_WIDTH, SCREEN_HEIGHT,
                                    OLED_MOSI_7_8, OLED_CLK_7_8, OLED_DC_7_8, OLED_reset_7_8, OLED_CS_7_8);


// --------------------------- Display interface ---------------------------- //
// ========================================================================== //

#if defined(ADAFRUIT_SSD1306_HAS_SETBUFFER) && ADAFRUIT_SSD1306_HAS_SETBUFFER
// We'll use a static buffer to avoid dynamic memory usage, and to allow
// multiple displays to reuse one single buffer.
static uint8_t buffer[(SCREEN_WIDTH * SCREEN_HEIGHT + 7) / 8];
#endif


// Implement the display interface, specifically, the begin and drawBackground
// methods.
class MySSD1306_DisplayInterface : public SSD1306_DisplayInterface {
  public:
    MySSD1306_DisplayInterface(Adafruit_SSD1306 &display)
      : SSD1306_DisplayInterface(display) {}

    void begin() override {
#if defined(ADAFRUIT_SSD1306_HAS_SETBUFFER) && ADAFRUIT_SSD1306_HAS_SETBUFFER
      disp.setBuffer(buffer);
#endif
      // Initialize the Adafruit_SSD1306 display
      if (!disp.begin())
        FATAL_ERROR(F("SSD1306 initialization failed."), 0x1306);

      // If you override the begin method, remember to call the super class method
      SSD1306_DisplayInterface::begin();
    }

    void drawBackground() override {
      disp.drawLine(1, 8, 126, 8, WHITE);
    }

} display_1_2 = ssd1306Display_1_2, display_3_4 = ssd1306Display_3_4, display_5_6 = ssd1306Display_5_6, display_7_8 = ssd1306Display_7_8;


// ------------------------------- Bank setup ------------------------------- //
// ========================================================================== //
/*
   Create a bank and a bank selector to change its setting.
*/

Bank<1> bank(8); // Create a new bank with two tracks per bank

// -------------------------- MIDI Input Elements --------------------------- //
// ========================================================================== //
/*
   Define all elements that listen for MIDI messages.
*/

// Main MCU LCD screen, used to get track names
MCU::LCD<> lcd {};

// Time display keeps track of the bar counter
MCU::TimeDisplay timedisplay {};

// Play / Record
NoteValue play {MCU::PLAY};
NoteValue record {MCU::RECORD};

// Mute
Bankable::NoteValue<1> mute[] {
  {bank, MCU::MUTE_1},
  {bank, MCU::MUTE_2},
  {bank, MCU::MUTE_3},
  {bank, MCU::MUTE_4},
  {bank, MCU::MUTE_5},
  {bank, MCU::MUTE_6},
  {bank, MCU::MUTE_7},
  {bank, MCU::MUTE_8},
};

// Solo
Bankable::NoteValue<1> solo[] {
  {bank, MCU::SOLO_1},
  {bank, MCU::SOLO_2},
  {bank, MCU::SOLO_3},
  {bank, MCU::SOLO_4},
  {bank, MCU::SOLO_5},
  {bank, MCU::SOLO_6},
  {bank, MCU::SOLO_7},
  {bank, MCU::SOLO_8},
};

NoteValue rudeSolo {MCU::RUDE_SOLO};

// Record arm / ready
Bankable::NoteValue<1> recrdy[] {
  {bank, MCU::REC_RDY_1},
  {bank, MCU::REC_RDY_2},
  {bank, MCU::REC_RDY_3},
  {bank, MCU::REC_RDY_4},
  {bank, MCU::REC_RDY_5},
  {bank, MCU::REC_RDY_6},
  {bank, MCU::REC_RDY_7},
  {bank, MCU::REC_RDY_8},
};


// VU meters
MCU::Bankable::VU<1> vu[] {
  {bank, 1, MCU::VUDecay::Hold},
  {bank, 2, MCU::VUDecay::Hold},
  {bank, 3, MCU::VUDecay::Hold},
  {bank, 4, MCU::VUDecay::Hold},
  {bank, 5, MCU::VUDecay::Hold},
  {bank, 6, MCU::VUDecay::Hold},
  {bank, 7, MCU::VUDecay::Hold},
  {bank, 8, MCU::VUDecay::Hold},
};

// VPot rings
MCU::Bankable::VPotRing<1> vpot[] {
  {bank, 1},
  {bank, 2},
  {bank, 3},
  {bank, 4},
  {bank, 5},
  {bank, 6},
  {bank, 7},
  {bank, 8},
};


// ---------------------------- Display Elements ---------------------------- //
// ========================================================================== //
/*
   Define all display elements that display the state of the input elements.
*/

// Track names
MCU::LCDDisplay lcddisps[] {
  // track (1), position (0, 40), font size (1)
  {display_1_2, lcd, bank, 1, {0, 40}, 1, WHITE},
  {display_1_2, lcd, bank, 2, {64, 40}, 1, WHITE},
  {display_3_4, lcd, bank, 3, {0, 40}, 1, WHITE},
  {display_3_4, lcd, bank, 4, {64, 40}, 1, WHITE},
  {display_5_6, lcd, bank, 5, {0, 40}, 1, WHITE},
  {display_5_6, lcd, bank, 6, {64, 40}, 1, WHITE},
  {display_7_8, lcd, bank, 7, {0, 40}, 1, WHITE},
  {display_7_8, lcd, bank, 8, {64, 40}, 1, WHITE},
};

// Time display
MCU::TimeDisplayDisplay timedisplaydisplay {
  // position (0, 0), font size (1)
  display_1_2, timedisplay, {0, 0}, 1, WHITE,
};

// Play / Record
BitmapDisplay<> playDisp {
  display_1_2, play, XBM::play_7, {16 + 64, 0}, WHITE,
};
BitmapDisplay<> recordDisp {
  display_1_2, record, XBM::record_7, {26 + 64, 0}, WHITE,
};

// Mute
BitmapDisplay<> muteDisp[] {
  {display_1_2, mute[0], XBM::mute_10B, {14, 50}, WHITE},
  {display_1_2, mute[1], XBM::mute_10B, {14 + 64, 50}, WHITE},
  {display_3_4, mute[2], XBM::mute_10B, {14, 50}, WHITE},
  {display_3_4, mute[3], XBM::mute_10B, {14 + 64, 50}, WHITE},
  {display_5_6, mute[4], XBM::mute_10B, {14, 50}, WHITE},
  {display_5_6, mute[5], XBM::mute_10B, {14 + 64, 50}, WHITE},
  {display_7_8, mute[6], XBM::mute_10B, {14, 50}, WHITE},
  {display_7_8, mute[7], XBM::mute_10B, {14 + 64, 50}, WHITE},
};

// Solo
BitmapDisplay<> soloDisp[] {
  {display_1_2, solo[0], XBM::solo_10B, {14, 50}, WHITE},
  {display_1_2, solo[1], XBM::solo_10B, {14 + 64, 50}, WHITE},
  {display_3_4, solo[2], XBM::solo_10B, {14, 50}, WHITE},
  {display_3_4, solo[3], XBM::solo_10B, {14 + 64, 50}, WHITE},
  {display_5_6, solo[4], XBM::solo_10B, {14, 50}, WHITE},
  {display_5_6, solo[5], XBM::solo_10B, {14 + 64, 50}, WHITE},
  {display_7_8, solo[6], XBM::solo_10B, {14, 50}, WHITE},
  {display_7_8, solo[7], XBM::solo_10B, {14 + 64, 50}, WHITE},
};

BitmapDisplay<> rudeSoloDisp {
  display_7_8, rudeSolo, XBM::solo_7, {36 + 64, 0}, WHITE};

// Record arm / ready
BitmapDisplay<> recrdyDisp[] {
  {display_1_2, recrdy[0], XBM::rec_rdy_10B, {14 + 14, 50}, WHITE},
  {display_1_2, recrdy[1], XBM::rec_rdy_10B, {14 + 14 + 64, 50}, WHITE},
  {display_3_4, recrdy[2], XBM::rec_rdy_10B, {14 + 14, 50}, WHITE},
  {display_3_4, recrdy[3], XBM::rec_rdy_10B, {14 + 14 + 64, 50}, WHITE},
  {display_5_6, recrdy[4], XBM::rec_rdy_10B, {14 + 14, 50}, WHITE},
  {display_5_6, recrdy[5], XBM::rec_rdy_10B, {14 + 14 + 64, 50}, WHITE},
  {display_7_8, recrdy[6], XBM::rec_rdy_10B, {14 + 14, 50}, WHITE},
  {display_7_8, recrdy[7], XBM::rec_rdy_10B, {14 + 14 + 64, 50}, WHITE},
};

// VU meters
MCU::VUDisplay<> vuDisp[] {
  // position (32+11, 60), width (16), bar height (3) px, bar spacing (1) px
  {display_1_2, vu[0], {32 + 11, 60}, 16, 3, 1, WHITE},
  {display_1_2, vu[1], {32 + 11 + 64, 60}, 16, 3, 1, WHITE},
  {display_3_4, vu[2], {32 + 11, 60}, 16, 3, 1, WHITE},
  {display_3_4, vu[3], {32 + 11 + 64, 60}, 16, 3, 1, WHITE},
  {display_5_6, vu[4], {32 + 11, 60}, 16, 3, 1, WHITE},
  {display_5_6, vu[5], {32 + 11 + 64, 60}, 16, 3, 1, WHITE},
  {display_7_8, vu[6], {32 + 11, 60}, 16, 3, 1, WHITE},
  {display_7_8, vu[7], {32 + 11 + 64, 60}, 16, 3, 1, WHITE},

};

// VPot rings
MCU::VPotDisplay<> vpotDisp[] {
  // position (0, 10), outer radius (14) px, inner radius (12) px
  {display_1_2, vpot[0], {0, 10}, 14, 12, WHITE},
  {display_1_2, vpot[1], {64, 10}, 14, 12, WHITE},
  {display_3_4, vpot[2], {0, 10}, 14, 12, WHITE},
  {display_3_4, vpot[3], {64, 10}, 14, 12, WHITE},
  {display_5_6, vpot[4], {0, 10}, 14, 12, WHITE},
  {display_5_6, vpot[5], {64, 10}, 14, 12, WHITE},
  {display_7_8, vpot[6], {0, 10}, 14, 12, WHITE},
  {display_7_8, vpot[7], {64, 10}, 14, 12, WHITE},
};

// Bank seting
BankDisplay bankDisp[] {
  // first track of the bank (1), position (0, 50), font size (2)
  {display_1_2, bank, 1, {0, 50}, 2, WHITE},
  {display_1_2, bank, 2, {64, 50}, 2, WHITE},
  {display_3_4, bank, 3, {0, 50}, 2, WHITE},
  {display_3_4, bank, 4, {64, 50}, 2, WHITE},
  {display_5_6, bank, 5, {0, 50}, 2, WHITE},
  {display_5_6, bank, 6, {64, 50}, 2, WHITE},
  {display_7_8, bank, 7, {0, 50}, 2, WHITE},
  {display_7_8, bank, 8, {64, 50}, 2, WHITE},
};

// --------------------------Rotary Encoders (x16)--------------------------- //
// ========================================================================== //
//Track PAN Control Change Rotary Encoders (x8)

// Type for the MCP23017 encoders (translates encoder pulses to position)
using WireType = decltype(Wire);     // The type of I²C driver to use
using EncoderPositionType = uint8_t; // The type for saving encoder positions
using MCPEncoderType = MCP23017Encoders<WireType, EncoderPositionType>;

// Type for the MIDI encoders (translates position to MIDI messages)
struct CCMCPEncoder : GenericMIDIRotaryEncoder<MCPEncoderType::MCP23017Encoder,
  RelativeCCSender> {
  CCMCPEncoder(MCPEncoderType::MCP23017Encoder enc, MIDIAddress address,
               int16_t multiplier = 4, uint8_t pulsesPerStep = 4)
    : GenericMIDIRotaryEncoder(std::move(enc), address, multiplier,
                               pulsesPerStep, {}) {}
};

// Type for the MIDI encoders (translates position to MIDI messages)
struct PBMCPEncoder : GenericMIDIAbsoluteEncoder<MCPEncoderType::MCP23017Encoder,
  PitchBendSender<14>> {
  PBMCPEncoder(MCPEncoderType::MCP23017Encoder enc, MIDIAddress address,
               int16_t multiplier = 512, uint8_t pulsesPerStep = 4)
    : GenericMIDIAbsoluteEncoder(std::move(enc), address, multiplier,
                                 pulsesPerStep, {}) {}
};


// Create an object that manages the 8 encoders connected to the MCP23017.
MCPEncoderType encPAN {Wire, 0x20, 19};
//                  │     │    └─ Interrupt pin
//                  │     └────── Address offset
//                  └──────────── I²C interface

// Create an object that manages the 8 encoders connected to the MCP23017.
MCPEncoderType encVOL {Wire, 0x22, 5};
//                  │     │    └─ Interrupt pin
//                  │     └────── Address offset
//                  └──────────── I²C interface

// Instantiate 8 MIDI rotary encoders.
CCMCPEncoder ccencodersPAN[] {
  {
    encPAN[0],       // The encoder to use
    MCU::V_POT_1, // The MIDI address
    6,            // Encoder speed multiplier
    3,            // Number of pulses per physical "click" of the encoder
  },
  { encPAN[1], MCU::V_POT_2, // The MIDI address
    6,            // Encoder speed multiplier
    3,            // Number of pulses per physical "click" of the encoder},
  },
  { encPAN[2], MCU::V_POT_3, // The MIDI address
    6,            // Encoder speed multiplier
    3,            // Number of pulses per physical "click" of the encoder},
  },
  { encPAN[3], MCU::V_POT_4, // The MIDI address
    6,            // Encoder speed multiplier
    3,            // Number of pulses per physical "click" of the encoder},
  },
  { encPAN[4], MCU::V_POT_5, // The MIDI address
    -6,            // Encoder speed multiplier ***NEGATIVE VALUES INVERT THE DIRECTION OF ENCODER READING***
    3,            // Number of pulses per physical "click" of the encoder},
  },
  { encPAN[5], MCU::V_POT_6, // The MIDI address
    -6,            // Encoder speed multiplier ***NEGATIVE VALUES INVERT THE DIRECTION OF ENCODER READING***
    3,            // Number of pulses per physical "click" of the encoder},
  },
  { encPAN[6], MCU::V_POT_7, // The MIDI address
    -6,            // Encoder speed multiplier ***NEGATIVE VALUES INVERT THE DIRECTION OF ENCODER READING***
    3,            // Number of pulses per physical "click" of the encoder},
  },
  { encPAN[7], MCU::V_POT_8, // The MIDI address
    -6,            // Encoder speed multiplier ***NEGATIVE VALUES INVERT THE DIRECTION OF ENCODER READING***
    3,            // Number of pulses per physical "click" of the encoder},
  }
};

// Instantiate 8 MIDI rotary encoders.
PBMCPEncoder pbencodersVOL[] {
  { encVOL[0], CHANNEL_1, // The MIDI address
    512,            // Encoder speed multiplier
    4,            // Number of pulses per physical "click" of the encoder},
  },
  { encVOL[1], CHANNEL_2, // The MIDI address
    512,            // Encoder speed multiplier
    4,            // Number of pulses per physical "click" of the encoder},
  },
  { encVOL[2], CHANNEL_3, // The MIDI address
    512,            // Encoder speed multiplier
    4,            // Number of pulses per physical "click" of the encoder},
  },
  { encVOL[3], CHANNEL_4, // The MIDI address
    512,            // Encoder speed multiplier
    4,            // Number of pulses per physical "click" of the encoder},
  },
  { encVOL[4], CHANNEL_5, // The MIDI address
    -512,            // Encoder speed multiplier ***NEGATIVE VALUES INVERT THE DIRECTION OF ENCODER READING***
    4,            // Number of pulses per physical "click" of the encoder},
  },
  { encVOL[5], CHANNEL_6, // The MIDI address
    -512,            // Encoder speed multiplier ***NEGATIVE VALUES INVERT THE DIRECTION OF ENCODER READING***
    4,            // Number of pulses per physical "click" of the encoder},
  },
  { encVOL[6], CHANNEL_7, // The MIDI address
    -512,            // Encoder speed multiplier ***NEGATIVE VALUES INVERT THE DIRECTION OF ENCODER READING***
    4,            // Number of pulses per physical "click" of the encoder},
  },
  { encVOL[7], CHANNEL_8, // The MIDI address
    -512,            // Encoder speed multiplier ***NEGATIVE VALUES INVERT THE DIRECTION OF ENCODER READING***
    4,            // Number of pulses per physical "click" of the encoder},
  }
};

// ------------------------ Setup Buttons and Cursor------------------------- //
// ========================================================================== //

//Butons - 16 Analog IO Multiplexers HC4067 (x4)
CD74HC4067 mux1 {
  A0,              // Analog input pin
  {14, 15, 16, 17},// Address pins S0, S1, S2, S3
  22                // Enable Pin
};

CD74HC4067 mux2 {
  A0,              // Analog input pin
  {14, 15, 16, 17},// Address pins S0, S1, S2, S3
  23                // Enable Pin
};

CD74HC4067 mux3 {
  A0,              // Analog input pin
  {14, 15, 16, 17},// Address pins S0, S1, S2, S3
  24                // Enable Pin
};

NoteButton buttons1[] {
  {mux1.pin(0),  74 , 127},          //Arrangement/Session View
  {mux1.pin(1),  75 , 127},          //Clip/Devive View
  {mux1.pin(2),  77, 127},           //Show/Hide Browser
  {mux1.pin(3),  78, 127},           //Show/Hide INFO View
  {mux1.pin(4),  89, 127},           //Timeline Begining
  {mux1.pin(5),  MCU::ZOOM, 127},    //Play Selected Clip
  {mux1.pin(6),  MCU::SCRUB, 127},   //Play All Clips (Scene)
  {mux1.pin(7),  76, 127},           //Undo
  {mux1.pin(8),  MCU::REC_RDY_1, 127},
  {mux1.pin(9),  MCU::REC_RDY_2, 127},
  {mux1.pin(10), MCU::REC_RDY_3, 127},
  {mux1.pin(11), MCU::REC_RDY_4, 127},
  {mux1.pin(12), MCU::REC_RDY_5, 127},
  {mux1.pin(13), MCU::REC_RDY_6, 127},
  {mux1.pin(14), MCU::REC_RDY_7, 127},
  {mux1.pin(15), MCU::REC_RDY_8, 127},
};

NoteButton buttons2[] {
  {mux2.pin(0), 88, 127}, //If you change the last number "127" to "0"
  {mux2.pin(1), MCU::CYCLE, 127}, //you send a Note Off Message
  {mux2.pin(2), 87, 127},
  {mux2.pin(3), MCU::FAST_FWD, 127},
  {mux2.pin(4), MCU::REWIND, 127},
  {mux2.pin(5), MCU::RECORD, 127},
  {mux2.pin(6), MCU::STOP, 127},
  {mux2.pin(7), MCU::PLAY, 127},
  {mux2.pin(8), MCU::SELECT_8, 127},
  {mux2.pin(9), MCU::SELECT_7, 127},
  {mux2.pin(10), MCU::SELECT_6, 127},
  {mux2.pin(11), MCU::DOWN, 127},
  {mux2.pin(12), MCU::UP, 127},
  {mux2.pin(13), MCU::RIGHT, 127},
  {mux2.pin(14), MCU::LEFT, 127},
  {mux2.pin(15), MCU::SCRUB, 127},
};

NoteButton buttons3[] {
  {mux3.pin(0), MCU::SOLO_8},      //Midi Note 96 = Cursor Move Up
  {mux3.pin(1), MCU::SOLO_7},      //Midi Note 97 = Cursor Move Down
  {mux3.pin(2), MCU::SOLO_6},      //Midi Note 98 = Cursor Move Left
  {mux3.pin(3), MCU::SOLO_5},      //Midi Note 99 = Cursor Move Right
  {mux3.pin(4), MCU::SOLO_4},      //Midi Note 100 = Cursor Center Function
  {mux3.pin(5), MCU::SOLO_3},      //Midi Note 74 = Session View/Arrangement View
  {mux3.pin(6), MCU::SOLO_2},      //Midi Note 77 = Hide/Show Browser Side Bar
  {mux3.pin(7), MCU::SOLO_1},      //Midi Note 78 = Show/Hide Info/Detail View Window
  {mux3.pin(8), MCU::MUTE_1},      //Midi Note 75 = Clip View Selector Clip Preview/Device Preview
  {mux3.pin(9), MCU::MUTE_2},      //Midi Note 76 = Undo
  {mux3.pin(10), MCU::MUTE_3},     //MCU::SCRUB = Plays all the Clips of the Session and then the Next row of Clips
  {mux3.pin(11), MCU::MUTE_4},     //Midi Note  83 = Draw Mode
  {mux3.pin(12), MCU::MUTE_5},     //Midi Note 82 = Αdd Locator
  {mux3.pin(13), MCU::MUTE_6},     //Midi Note 89 = Timeline Start
  {mux3.pin(14), MCU::MUTE_7},     //Midi Note 90 = Timeline End
  {mux3.pin(15), MCU::MUTE_8},     //PLAY SCENE 101
};

NoteButton buttons4[] {
  //  {mux1.pin(0), 75},             //Midi Note 96 = Cursor Move Up
  //  {mux1.pin(1), 76},             //Midi Note 97 = Cursor Move Down
  //  {mux1.pin(2), MCU::SCRUB},     //Midi Note 98 = Cursor Move Left
  //  {mux1.pin(3), MCU::FAST_FWD},  //Midi Note 99 = Cursor Move Right
  //  {mux1.pin(4), MCU::REWIND},    //Midi Note 100 = Cursor Center Function
  //  {mux1.pin(5), MCU::RECORD},    //Midi Note 74 = Session View/Arrangement View
  //  {mux1.pin(6), MCU::STOP},      //Midi Note 77 = Hide/Show Browser Side Bar
  //  {mux1.pin(7), MCU::PLAY},      //Midi Note 78 = Show/Hide Info/Detail View Window
  {mux1.pin(8), MCU::SELECT_1},      //Midi Note 75 = Clip View Selector Clip Preview/Device Preview
  {mux1.pin(9), MCU::SELECT_2},      //Midi Note 76 = Undo
  {mux1.pin(10), MCU::SELECT_3},     //MCU::SCRUB = Plays all the Clips of the Session and then the Next row of Clips
  {mux1.pin(11), MCU::SELECT_4},     //Midi Note  83 = Draw Mode
  {mux1.pin(12), MCU::SELECT_5},     //Midi Note 82 = Αdd Locator
  {mux1.pin(13), MCU::SELECT_6},     //Midi Note 89 = Timeline Start
  {mux1.pin(14), MCU::SELECT_7},     //Midi Note 90 = Timeline End
  {mux1.pin(15), MCU::SELECT_8},     //PLAY SCENE 101
};

// ------------------------------ LEDs Setup -------------------------------- //
// ========================================================================== //
// Define the array of leds.

Array<CRGB, 8> leds {};
//The data pin with the strip connected.
constexpr uint8_t ledpin = 18;

// Create a functor that maps the velocity and the index of a note to a color.
struct RainbowColorMapper {
  CHSV operator()(uint8_t velocity, uint8_t index) const {
    return CHSV(255 * index / leds.length, 255, 255u * velocity / 127u);
  }
};

NoteRangeFastLED<leds.length, RainbowColorMapper> midiled {
  leds,
  MIDI_Notes::C(1),
};

// ----------------------- Master Volume Fader Setup ------------------------ //
// ========================================================================== //

PBPotentiometer potentiometer {
  A1,        // Analog pin connected to potentiometer
  CHANNEL_9, // MIDI Channel 1
};

void setup() {
  Control_Surface.begin();

  Wire.begin(); // Must be called before enc.begin()
  Wire.setClock(800000);

  encPAN.begin(); // Initialize the MCP23017
  encVOL.begin(); // Initialize the MCP23017
  RelativeCCSender::setMode(relativeCCmode::MACKIE_CONTROL_RELATIVE);

  ssd1306Display_1_2.clearDisplay();
  ssd1306Display_3_4.clearDisplay();
  ssd1306Display_5_6.clearDisplay();
  ssd1306Display_7_8.clearDisplay();

  // See FastLED examples and documentation for more information.
  FastLED.addLeds<NEOPIXEL, ledpin>(leds.data, leds.length);
  FastLED.setCorrection(TypicalPixelString);
  midiled.setBrightness(128);

  delay(200);

}

void loop() {
  Control_Surface.loop();

  ssd1306Display_1_2.clearDisplay();
  ssd1306Display_3_4.clearDisplay();
  ssd1306Display_5_6.clearDisplay();
  ssd1306Display_7_8.clearDisplay();

  encPAN.update();
  encVOL.update();

  if (midiled.getDirty()) { // If the colors changed
    FastLED.show();         // Update the LEDs with the new colors
  }
}
